# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the fourier module.
"""

import numpy as np
import pytest
from astropy.modeling.fitting import TRFLSQFitter
from astropy.modeling.models import Gaussian2D
from astropy.utils.exceptions import AstropyDeprecationWarning
from numpy.testing import assert_allclose

from photutils.psf_matching.fourier import (create_matching_kernel,
                                            make_kernel, make_wiener_kernel)
from photutils.psf_matching.tests.conftest import (
    _make_gaussian_psf, _make_gaussian_psf_noncentered)
from photutils.psf_matching.windows import SplitCosineBellWindow


class TestMakeKernel:
    def test_with_window(self, psf1, psf2):
        """
        Test with noiseless 2D Gaussians and a window.
        """
        size = psf1.shape[0]
        cen = (size - 1) / 2.0
        yy, xx = np.mgrid[0:size, 0:size]
        window = SplitCosineBellWindow(0.0, 0.2)
        kernel = make_kernel(psf1, psf2, window=window)

        fitter = TRFLSQFitter()
        gm1 = Gaussian2D(1.0, cen, cen, 3.0, 3.0)
        gfit = fitter(gm1, xx, yy, kernel)
        assert_allclose(gfit.x_stddev, gfit.y_stddev)
        assert_allclose(gfit.x_stddev, np.sqrt(25 - 9), atol=0.06)

    def test_without_window(self, psf1, psf2):
        """
        Test without a window function.
        """
        kernel = make_kernel(psf1, psf2)
        assert kernel.shape == psf1.shape
        assert_allclose(kernel.sum(), 1.0)

    def test_shape_mismatch(self, psf1):
        """
        Test that mismatched PSF shapes raise ValueError.
        """
        psf_small = _make_gaussian_psf(5, 1.5)
        match = 'must have the same shape'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, psf_small)

    def test_non_2d_source(self, psf2):
        """
        Test that non-2D source PSF raises ValueError.
        """
        match = 'source_psf must be a 2D array'
        with pytest.raises(ValueError, match=match):
            make_kernel(np.ones(25), psf2)

    def test_non_2d_target(self, psf1):
        """
        Test that non-2D target PSF raises ValueError.
        """
        match = 'target_psf must be a 2D array'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, np.ones(25))

    def test_even_shape(self):
        """
        Test that even-shaped PSFs raise ValueError.
        """
        psf = np.zeros((4, 4))
        psf[2, 2] = 1.0
        match = 'must have odd dimensions'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf, psf)

    def test_non_callable_window(self, psf1, psf2):
        """
        Test that non-callable window raises TypeError.
        """
        match = 'window must be a callable'
        with pytest.raises(TypeError, match=match):
            make_kernel(psf1, psf2, window='bad')

    def test_regularization(self, psf1, psf2):
        """
        Test with an aggressive regularization threshold.
        """
        kernel = make_kernel(psf1, psf2, regularization=0.5)
        assert kernel.shape == psf1.shape
        assert_allclose(kernel.sum(), 1.0)

    def test_regularization_zero(self, psf1, psf2):
        """
        Test with regularization=0 (minimum thresholding).
        """
        kernel = make_kernel(psf1, psf2, regularization=0)
        assert kernel.shape == psf1.shape
        assert_allclose(kernel.sum(), 1.0)

    def test_regularization_negative(self, psf1, psf2):
        """
        Test that negative regularization raises an error.
        """
        match = 'regularization must be in the range'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, psf2, regularization=-0.1)

    def test_regularization_greater_than_one(self, psf1, psf2):
        """
        Test that regularization > 1 raises an error.
        """
        match = 'regularization must be in the range'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, psf2, regularization=1.5)

    def test_regularization_invalid(self, psf1, psf2):
        """
        Test that regularization=1 raises an error (range is [0, 1)).
        """
        match = 'regularization must be in the range'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, psf2, regularization=1.0)

    def test_window_not_2d(self, psf1, psf2):
        """
        Test that window function returning non-2D array raises error.
        """
        def bad_window(shape):
            return np.ones(shape[0])  # 1D array

        match = 'window function must return a 2D array'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, psf2, window=bad_window)

    def test_window_wrong_shape(self, psf1, psf2):
        """
        Test that window function returning wrong shape raises error.
        """
        def bad_window(shape):  # noqa: ARG001
            return np.ones((10, 10))  # wrong shape

        match = 'window function must return an array with shape'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, psf2, window=bad_window)

    def test_window_values_below_zero(self, psf1, psf2):
        """
        Test that window function with values < 0 raises error.
        """
        def bad_window(shape):
            arr = np.ones(shape)
            arr[0, 0] = -0.1
            return arr

        match = 'window function values must be in the range'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, psf2, window=bad_window)

    def test_window_values_above_one(self, psf1, psf2):
        """
        Test that window function with values > 1 raises error.
        """
        def bad_window(shape):
            arr = np.ones(shape)
            arr[0, 0] = 1.5
            return arr

        match = 'window function values must be in the range'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, psf2, window=bad_window)

    def test_asymmetric_shape(self):
        """
        Test with asymmetric PSF shapes.
        """
        # Create 51x25 PSFs centered at (x=12, y=25)
        shape = (51, 25)
        psf1 = _make_gaussian_psf_noncentered(shape, 3, xcen=12, ycen=25)
        psf2 = _make_gaussian_psf_noncentered(shape, 5, xcen=12, ycen=25)

        kernel = make_kernel(psf1, psf2)
        assert kernel.shape == shape
        assert_allclose(kernel.sum(), 1.0)

    def test_zero_kernel_raises(self, psf1, psf2):
        """
        Test that a window returning all zeros raises ValueError because
        the computed kernel sums to zero.
        """
        def zero_window(shape):
            return np.zeros(shape)

        match = 'The computed kernel sums to zero'
        with pytest.raises(ValueError, match=match):
            make_kernel(psf1, psf2, window=zero_window)


class TestMakeKernelWiener:
    def test_basic(self, psf1, psf2):
        """
        Test basic Wiener kernel creation with noiseless Gaussians.
        """
        kernel = make_wiener_kernel(psf1, psf2)
        assert kernel.shape == psf1.shape
        assert_allclose(kernel.sum(), 1.0)

    def test_kernel_shape(self, psf1, psf2):
        """
        Test that the kernel has the expected Gaussian shape.

        For two Gaussians with sigma=3 and sigma=5, the matching
        kernel should be a Gaussian with sigma=sqrt(25-9)=4.
        """
        size = psf1.shape[0]
        cen = (size - 1) / 2.0
        yy, xx = np.mgrid[0:size, 0:size]
        kernel = make_wiener_kernel(psf1, psf2)

        fitter = TRFLSQFitter()
        gm1 = Gaussian2D(1.0, cen, cen, 3.0, 3.0)
        gfit = fitter(gm1, xx, yy, kernel)
        assert_allclose(gfit.x_stddev, gfit.y_stddev)
        assert_allclose(gfit.x_stddev, np.sqrt(25 - 9), atol=0.06)

    def test_with_window(self, psf1, psf2):
        """
        Test with a window function applied.
        """
        window = SplitCosineBellWindow(0.0, 0.2)
        kernel = make_wiener_kernel(psf1, psf2, window=window)
        assert kernel.shape == psf1.shape
        assert_allclose(kernel.sum(), 1.0)

    def test_regularization(self, psf1, psf2):
        """
        Test with different regularization strengths.
        """
        kernel_weak = make_wiener_kernel(psf1, psf2, regularization=1e-8)
        kernel_strong = make_wiener_kernel(psf1, psf2, regularization=1e-1)
        # Both should be normalized
        assert_allclose(kernel_weak.sum(), 1.0)
        assert_allclose(kernel_strong.sum(), 1.0)
        # Stronger regularization should produce a smoother kernel
        # (lower max value)
        assert kernel_strong.max() < kernel_weak.max()

    def test_shape_mismatch(self, psf1):
        """
        Test that mismatched PSF shapes raise ValueError.
        """
        g_small = _make_gaussian_psf(5, 1.5)
        match = 'must have the same shape'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, g_small)

    def test_non_2d_source(self, psf2):
        """
        Test that non-2D source PSF raises ValueError.
        """
        match = 'source_psf must be a 2D array'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(np.ones(25), psf2)

    def test_non_2d_target(self, psf1):
        """
        Test that non-2D target PSF raises ValueError.
        """
        match = 'target_psf must be a 2D array'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, np.ones(25))

    def test_even_shape(self):
        """
        Test that even-shaped PSFs raise ValueError.
        """
        psf = np.zeros((4, 4))
        psf[2, 2] = 1.0
        match = 'must have odd dimensions'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf, psf)

    def test_non_callable_window(self, psf1, psf2):
        """
        Test that non-callable window raises TypeError.
        """
        match = 'window must be a callable'
        with pytest.raises(TypeError, match=match):
            make_wiener_kernel(psf1, psf2, window='bad')

    def test_negative_regularization(self, psf1, psf2):
        """
        Test that negative regularization raises ValueError.
        """
        match = 'regularization must be a positive number'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, regularization=-1.0)

    def test_zero_regularization(self, psf1, psf2):
        """
        Test that zero regularization raises ValueError.
        """
        match = 'regularization must be a positive number'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, regularization=0.0)

    def test_penalty_string_equals_array(self, psf1, psf2):
        """
        Test that 'laplacian' string gives the same result as the
        explicit Laplacian array.
        """
        laplacian = np.array([[0, -1, 0],
                              [-1, 4, -1],
                              [0, -1, 0]])
        kernel_str = make_wiener_kernel(psf1, psf2, penalty='laplacian')
        kernel_arr = make_wiener_kernel(psf1, psf2, penalty=laplacian)
        assert_allclose(kernel_str, kernel_arr)

    def test_penalty_laplacian_kernel_shape(self, psf1, psf2):
        """
        Test that Laplacian penalty kernel has the expected Gaussian
        shape.
        """
        size = psf1.shape[0]
        cen = (size - 1) / 2.0
        yy, xx = np.mgrid[0:size, 0:size]
        kernel = make_wiener_kernel(psf1, psf2, penalty='laplacian')

        fitter = TRFLSQFitter()
        gm1 = Gaussian2D(1.0, cen, cen, 3.0, 3.0)
        gfit = fitter(gm1, xx, yy, kernel)
        assert_allclose(gfit.x_stddev, gfit.y_stddev)
        assert_allclose(gfit.x_stddev, np.sqrt(25 - 9), atol=0.06)

    def test_penalty_invalid_string(self, psf1, psf2):
        """
        Test that an invalid penalty string raises ValueError.
        """
        match = 'Invalid penalty string'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, penalty='invalid')

    def test_penalty_invalid_type(self, psf1, psf2):
        """
        Test that an invalid penalty type raises ValueError.
        """
        match = 'penalty must be None'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, penalty=42)

    def test_penalty_non_2d_array(self, psf1, psf2):
        """
        Test that a non-2D penalty array raises ValueError.
        """
        match = 'penalty array must be 2D'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, penalty=np.ones(5))

    def test_penalty_psf_too_small_for_laplacian(self):
        """
        Test that a PSF smaller than 3x3 raises ValueError when using
        laplacian penalty.
        """
        # Create 1x1 PSFs (too small for 3x3 laplacian)
        psf1 = np.array([[1.0]])
        psf2 = np.array([[1.0]])
        match = 'PSFs must be at least as large as the penalty operator'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, penalty='laplacian')

    def test_penalty_psf_too_small_for_biharmonic(self):
        """
        Test that a PSF smaller than 5x5 raises ValueError when using
        biharmonic penalty.
        """
        # Create 3x3 PSFs (too small for 5x5 biharmonic)
        psf1 = _make_gaussian_psf(3, 1.0)
        psf2 = _make_gaussian_psf(3, 1.5)
        match = 'PSFs must be at least as large as the penalty operator'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, penalty='biharmonic')

    def test_penalty_psf_minimum_size_laplacian(self):
        """
        Test that 3x3 PSF (minimum size) works with laplacian penalty.
        """
        psf1 = _make_gaussian_psf(3, 0.8)
        psf2 = _make_gaussian_psf(3, 1.0)
        kernel = make_wiener_kernel(psf1, psf2, penalty='laplacian')
        assert kernel.shape == (3, 3)
        assert_allclose(kernel.sum(), 1.0)

    def test_penalty_psf_minimum_size_biharmonic(self):
        """
        Test that 5x5 PSF (minimum size) works with biharmonic penalty.
        """
        psf1 = _make_gaussian_psf(5, 1.2)
        psf2 = _make_gaussian_psf(5, 1.5)
        kernel = make_wiener_kernel(psf1, psf2, penalty='biharmonic')
        assert kernel.shape == (5, 5)
        assert_allclose(kernel.sum(), 1.0)

    def test_penalty_custom_array_too_large(self):
        """
        Test that a custom penalty array larger than the PSF raises
        ValueError.
        """
        # Create 3x3 PSFs but 5x5 penalty
        psf1 = _make_gaussian_psf(3, 0.8)
        psf2 = _make_gaussian_psf(3, 1.0)
        penalty = np.ones((5, 5))
        match = 'PSFs must be at least as large as the penalty operator'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, penalty=penalty)

    def test_penalty_custom_array(self, psf1, psf2):
        """
        Test with a custom 2D penalty array.
        """
        # Use a simple high-pass operator
        penalty = np.array([[0, 0, 0],
                            [0, 1, 0],
                            [0, 0, 0]])
        kernel = make_wiener_kernel(psf1, psf2, penalty=penalty)
        assert kernel.shape == psf1.shape
        assert_allclose(kernel.sum(), 1.0)

    def test_penalty_with_window(self, psf1, psf2):
        """
        Test that penalty and window can be used together.
        """
        window = SplitCosineBellWindow(0.0, 0.2)
        kernel = make_wiener_kernel(psf1, psf2, penalty='laplacian',
                                    window=window)
        assert kernel.shape == psf1.shape
        assert_allclose(kernel.sum(), 1.0)

    def test_penalty_differs_from_scalar(self, psf1, psf2):
        """
        Test that Laplacian penalty gives a different result than scalar
        Tikhonov with the same regularization parameter.
        """
        reg = 1e-4
        kernel_scalar = make_wiener_kernel(psf1, psf2, regularization=reg)
        kernel_laplacian = make_wiener_kernel(psf1, psf2, regularization=reg,
                                              penalty='laplacian')
        assert not np.allclose(kernel_scalar, kernel_laplacian)

    def test_penalty_biharmonic_basic(self, psf1, psf2):
        """
        Test basic biharmonic penalty functionality.
        """
        kernel = make_wiener_kernel(psf1, psf2, penalty='biharmonic')
        assert kernel.shape == psf1.shape
        assert_allclose(kernel.sum(), 1.0)

    def test_penalty_biharmonic_kernel_shape(self, psf1, psf2):
        """
        Test that biharmonic penalty kernel has the expected Gaussian
        shape.
        """
        size = psf1.shape[0]
        cen = (size - 1) / 2.0
        yy, xx = np.mgrid[0:size, 0:size]
        kernel = make_wiener_kernel(psf1, psf2, penalty='biharmonic')

        fitter = TRFLSQFitter()
        gm1 = Gaussian2D(1.0, cen, cen, 3.0, 3.0)
        gfit = fitter(gm1, xx, yy, kernel)
        assert_allclose(gfit.x_stddev, gfit.y_stddev)
        assert_allclose(gfit.x_stddev, np.sqrt(25 - 9), atol=0.08)

    def test_penalty_biharmonic_results(self, psf1, psf2):
        """
        Test that biharmonic penalty gives different results than scalar
        and Laplacian penalties.
        """
        reg = 1e-4
        kernel_scalar = make_wiener_kernel(psf1, psf2, regularization=reg)
        kernel_laplacian = make_wiener_kernel(psf1, psf2, regularization=reg,
                                              penalty='laplacian')
        kernel_biharmonic = make_wiener_kernel(psf1, psf2, regularization=reg,
                                               penalty='biharmonic')
        assert not np.allclose(kernel_scalar, kernel_biharmonic)
        assert not np.allclose(kernel_laplacian, kernel_biharmonic)

    def test_penalty_biharmonic_smoothness(self, psf1, psf2):
        """
        Test that biharmonic penalty produces smoother kernels than
        Laplacian (lower peak value indicates more smoothing).
        """
        reg = 1e-4
        kernel_laplacian = make_wiener_kernel(psf1, psf2, regularization=reg,
                                              penalty='laplacian')
        kernel_biharmonic = make_wiener_kernel(psf1, psf2, regularization=reg,
                                               penalty='biharmonic')
        # Biharmonic should produce a smoother kernel with lower peak
        assert kernel_biharmonic.max() < kernel_laplacian.max()

    def test_penalty_biharmonic_with_window(self, psf1, psf2):
        """
        Test that biharmonic penalty works with window functions.
        """
        window = SplitCosineBellWindow(0.0, 0.2)
        kernel = make_wiener_kernel(psf1, psf2, penalty='biharmonic',
                                    window=window)
        assert kernel.shape == psf1.shape
        assert_allclose(kernel.sum(), 1.0)

    def test_asymmetric_shape(self):
        """
        Test with asymmetric PSF shapes.
        """
        # Create 51x25 PSFs centered at (x=12, y=25)
        shape = (51, 25)
        psf1 = _make_gaussian_psf_noncentered(shape, 3, xcen=12, ycen=25)
        psf2 = _make_gaussian_psf_noncentered(shape, 5, xcen=12, ycen=25)

        kernel = make_wiener_kernel(psf1, psf2)
        assert kernel.shape == shape
        assert_allclose(kernel.sum(), 1.0)

    def test_window_not_2d(self, psf1, psf2):
        """
        Test that window function returning non-2D array raises error.
        """
        def bad_window(shape):
            return np.ones(shape[0])  # 1D array

        match = 'window function must return a 2D array'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, window=bad_window)

    def test_window_wrong_shape(self, psf1, psf2):
        """
        Test that window function returning wrong shape raises error.
        """
        def bad_window(shape):  # noqa: ARG001
            return np.ones((10, 10))  # wrong shape

        match = 'window function must return an array with shape'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, window=bad_window)

    def test_window_values_below_zero(self, psf1, psf2):
        """
        Test that window function with values < 0 raises error.
        """
        def bad_window(shape):
            arr = np.ones(shape)
            arr[0, 0] = -0.1
            return arr

        match = 'window function values must be in the range'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, window=bad_window)

    def test_window_values_above_one(self, psf1, psf2):
        """
        Test that window function with values > 1 raises error.
        """
        def bad_window(shape):
            arr = np.ones(shape)
            arr[0, 0] = 1.5
            return arr

        match = 'window function values must be in the range'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, window=bad_window)

    def test_zero_kernel_raises(self, psf1, psf2):
        """
        Test that a window returning all zeros raises ValueError because
        the computed kernel sums to zero.
        """
        def zero_window(shape):
            return np.zeros(shape)

        match = 'The computed kernel sums to zero'
        with pytest.raises(ValueError, match=match):
            make_wiener_kernel(psf1, psf2, window=zero_window)


class TestCreateMatchingKernelDeprecated:
    def test_deprecation_warning(self, psf1, psf2):
        """
        Test that create_matching_kernel raises a deprecation warning.
        """
        with pytest.warns(AstropyDeprecationWarning):
            kernel = create_matching_kernel(psf1, psf2)
        assert_allclose(kernel.sum(), 1.0)

    def test_deprecation_result_matches_make_kernel(self, psf1, psf2):
        """
        Test that create_matching_kernel returns the same result as
        make_kernel.
        """
        with pytest.warns(AstropyDeprecationWarning):
            kernel_old = create_matching_kernel(psf1, psf2)
        kernel_new = make_kernel(psf1, psf2)
        assert_allclose(kernel_old, kernel_new)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Deprecated subpackage. Use ``photutils.psf_matching`` instead.
"""

import warnings

from astropy.utils.exceptions import AstropyDeprecationWarning

import photutils.psf_matching as _psf_matching
from photutils.psf_matching.fourier import __all__ as _fourier_all
from photutils.psf_matching.windows import __all__ as _windows_all

__all__ = list(_fourier_all) + list(_windows_all)

_deprecation_msg = ('photutils.psf.matching is deprecated (since version '
                    '3.0) and will be removed in a future version. Use '
                    'photutils.psf_matching instead. Please update your '
                    'imports accordingly.')


def __getattr__(name):
    if name in __all__:
        warnings.warn(_deprecation_msg, AstropyDeprecationWarning,
                      stacklevel=2)
        return getattr(_psf_matching, name)
    msg = f'module {__name__!r} has no attribute {name!r}'
    raise AttributeError(msg)


def __dir__():
    return __all__



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for optional dependencies.

Attributes ``HAS_<PKG>`` (e.g., ``HAS_MATPLOTLIB``, ``HAS_SKIMAGE``)
are booleans that indicate whether the corresponding package can be
imported. The actual import is performed lazily on first attribute
access via :pep:`562`.

The ``HAS_*`` names are derived from the *import* name of each optional
dependency (uppercased, hyphens/dots replaced by underscores).
For the handful of packages whose import name differs from their
distribution (pip) name, a small translation dict (``_DIST_TO_IMPORT``)
is maintained.
"""

import importlib
from importlib.metadata import packages_distributions, requires

from packaging.requirements import Requirement

# Hardcoded translation for packages whose import name differs from
# their distribution (pip) name. All other packages are assumed to be
# importable using their distribution name directly. If a new optional
# dependency is added whose import name does not match its dist name,
# add a single entry below.
_DIST_TO_IMPORT = {
    'scikit-image': 'skimage',
}


def _get_optional_deps(dist_name, *, extra='all'):
    """
    Return the optional-dependency distribution names for ``dist_name``.

    Parameters
    ----------
    dist_name : str
        The distribution (pip) name of the package whose metadata is
        queried (e.g., ``'photutils'``).

    extra : str, optional
        The extras group to query (default ``'all'``).

    Returns
    -------
    deps : list of str
        Sorted distribution names of the optional dependencies.
    """
    deps = set()
    for req_str in (requires(dist_name) or []):
        req = Requirement(req_str)
        if req.marker and req.marker.evaluate({'extra': extra}):
            deps.add(req.name)
    return sorted(deps)


def _dist_to_has_key(dist_name):
    """
    Convert a distribution name to the corresponding ``HAS_*`` key.

    For example, ``'scikit-image'`` -> ``'SKIMAGE'`` and
    ``'matplotlib'`` -> ``'MATPLOTLIB'``.
    """
    import_name = _DIST_TO_IMPORT.get(dist_name, dist_name)

    return import_name.upper().replace('-', '_').replace('.', '_')


# Derive the distribution name of this package from its top-level import
# name
_pkg_import_name = __name__.split('.')[0]
_pkg_dist_name = packages_distributions().get(_pkg_import_name,
                                              [_pkg_import_name])[0]

# Build lookup: HAS_* suffix -> dist name.
_optional_deps = _get_optional_deps(_pkg_dist_name, extra='all')
_deps_by_key = {_dist_to_has_key(d): d for d in _optional_deps}

__all__ = [f'HAS_{key}' for key in sorted(_deps_by_key)]

_cache = {}


# Implemented as a module-level __getattr__ to allow for lazy imports on
# first access. See PEP 562 for details.
def __getattr__(name):
    if name.startswith('HAS_'):
        key = name[4:]
        if key in _deps_by_key:
            if name not in _cache:
                dist_name = _deps_by_key[key]
                import_name = _DIST_TO_IMPORT.get(dist_name, dist_name)
                try:
                    importlib.import_module(import_name)
                    _cache[name] = True
                except ImportError:
                    _cache[name] = False
            return _cache[name]

    msg = f'Module {__name__!r} has no attribute {name!r}'
    raise AttributeError(msg)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for detecting sources in an image.
"""

from photutils.segmentation.deblend import deblend_sources
from photutils.segmentation.detect import detect_sources
from photutils.utils._deprecation import (deprecated_getattr,
                                          deprecated_positional_kwargs,
                                          deprecated_renamed_argument)
from photutils.utils._parameters import as_pair
from photutils.utils._repr import make_repr

__all__ = ['SourceFinder']

# Remove in 4.0
_FINDER_DEPRECATED_ATTRIBUTES = {
    'npixels': 'n_pixels',
    'nlevels': 'n_levels',
    'nproc': 'n_processes',
}


class SourceFinder:
    """
    Class to detect sources, including deblending, in an image using
    segmentation.

    This is a convenience class that combines the functionality
    of `~photutils.segmentation.detect_sources` and
    `~photutils.segmentation.deblend_sources`.

    Sources are deblended using a combination of
    multi-thresholding and `watershed segmentation
    <https://en.wikipedia.org/wiki/Watershed_(image_processing)>`_. In
    order to deblend sources, there must be a saddle between them.

    Parameters
    ----------
    n_pixels : int or array_like of 2 int
        The minimum number of connected pixels, each greater than a
        specified threshold, that an object must have to be detected. If
        ``n_pixels`` is an integer, then the value will be used for both
        source detection and deblending (which internally uses source
        detection at multiple thresholds). If ``n_pixels`` contains two
        values, then the first value will be used for source detection
        and the second value used for source deblending. ``n_pixels``
        values must be positive integers.

    connectivity : {4, 8}, optional
        The type of pixel connectivity used in determining how pixels
        are grouped into a detected source. The options are 4 or
        8 (default). 4-connected pixels touch along their edges.
        8-connected pixels touch along their edges or corners.

    deblend : bool, optional
        Whether to deblend overlapping sources.

    n_levels : int, optional
        The number of multi-thresholding levels to use for deblending.
        Each source will be re-thresholded at ``n_levels`` levels spaced
        exponentially or linearly (see the ``mode`` keyword) between
        its minimum and maximum values. This keyword is ignored unless
        ``deblend=True``.

    contrast : float, optional
        The fraction of the total source flux that a local peak must
        have (at any one of the multi-thresholds) to be deblended
        as a separate object. ``contrast`` must be between 0 and 1,
        inclusive. If ``contrast=0`` then every local peak will be made
        a separate object (maximum deblending). If ``contrast=1`` then
        no deblending will occur. The default is 0.001, which will
        deblend sources with a 7.5 magnitude difference. This keyword is
        ignored unless ``deblend=True``.

    mode : {'exponential', 'linear', 'sinh'}, optional
        The mode used in defining the spacing between the
        multi-thresholding levels (see the ``n_levels`` keyword)
        during deblending. The ``'exponential'`` and ``'sinh'`` modes
        have more threshold levels near the source minimum and less
        near the source maximum. The ``'linear'`` mode evenly spaces
        the threshold levels between the source minimum and maximum.
        The ``'exponential'`` and ``'sinh'`` modes differ in that
        the ``'exponential'`` levels are dependent on the source
        maximum/minimum ratio (smaller ratios are more linear; larger
        ratios are more exponential), while the ``'sinh'`` levels
        are not. Also, the ``'exponential'`` mode will be changed to
        ``'linear'`` for sources with non-positive minimum data values.
        This keyword is ignored unless ``deblend=True``.

    relabel : bool, optional
        If `True` (default), then the segmentation image will be
        relabeled after deblending such that the labels are in
        consecutive order starting from 1. This keyword is ignored
        unless ``deblend=True``.

    n_processes : int, optional
        The number of processes to use for source deblending. If set to
        1, then a serial implementation is used instead of a parallel
        one. If `None`, then the number of processes will be set to the
        number of CPUs detected on the machine. Please note that due to
        overheads, multiprocessing may be slower than serial processing
        if only a small number of sources are to be deblended. The
        benefits of multiprocessing require ~1000 or more sources to
        deblend, with larger gains as the number of sources increase.
        This keyword is ignored unless ``deblend=True``.

    progress_bar : bool, optional
        Whether to display a progress bar. If ``n_processes = 1``, then the
        ID shown after the progress bar is the source label being
        deblended. If multiprocessing is used (``n_processes > 1``), the
        ID shown is the last source label that was deblended. The
        progress bar requires that the `tqdm <https://tqdm.github.io/>`_
        optional dependency be installed. This keyword is ignored unless
        ``deblend=True``.

    See Also
    --------
    :func:`photutils.segmentation.detect_sources`
    :func:`photutils.segmentation.deblend_sources`

    Examples
    --------
    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from astropy.convolution import convolve
        from astropy.visualization import simple_norm
        from photutils.background import Background2D, MedianBackground
        from photutils.datasets import make_100gaussians_image
        from photutils.segmentation import (SourceFinder,
                                            make_2dgaussian_kernel)

        # Make a simulated image
        data = make_100gaussians_image()

        # Estimate the background using Background2D and subtract it
        bkg_estimator = MedianBackground()
        bkg = Background2D(data, (50, 50), filter_size=(3, 3),
                           bkg_estimator=bkg_estimator)
        data -= bkg.background  # subtract the background

        # Convolve the data
        kernel = make_2dgaussian_kernel(3.0, size=5)
        convolved_data = convolve(data, kernel)

        # Detect the sources
        threshold = 1.5 * bkg.background_rms  # per-pixel detection threshold
        finder = SourceFinder(n_pixels=10, progress_bar=False)
        segment_map = finder(convolved_data, threshold)

        # Plot the image and the segmentation image
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(8, 10))
        norm = simple_norm(data, 'sqrt', percent=99.5)
        ax1.imshow(data, norm=norm, origin='lower')
        segment_map.imshow(ax=ax2)
        fig.tight_layout()
    """

    @deprecated_renamed_argument('npixels', 'n_pixels', '3.0', until='4.0')
    @deprecated_renamed_argument('nlevels', 'n_levels', '3.0', until='4.0')
    @deprecated_renamed_argument('nproc', 'n_processes', '3.0', until='4.0')
    def __init__(self, n_pixels, *, connectivity=8, deblend=True, n_levels=32,
                 contrast=0.001, mode='exponential', relabel=True,
                 n_processes=1, progress_bar=True):
        self.n_pixels = as_pair('n_pixels', n_pixels, check_odd=False)
        self.deblend = deblend
        self.connectivity = connectivity
        self.n_levels = n_levels
        self.contrast = contrast
        self.mode = mode
        self.relabel = relabel
        self.n_processes = n_processes
        self.progress_bar = progress_bar

    def __repr__(self):
        params = ('n_pixels', 'deblend', 'connectivity', 'n_levels',
                  'contrast', 'mode', 'relabel', 'n_processes',
                  'progress_bar')
        return make_repr(self, params)

    # Remove in 4.0
    def __getattr__(self, name):
        return deprecated_getattr(self, name,
                                  _FINDER_DEPRECATED_ATTRIBUTES,
                                  since='3.0', until='4.0')

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __call__(self, data, threshold, mask=None):
        """
        Detect sources, including deblending, in an image using
        segmentation.

        Parameters
        ----------
        data : 2D `~numpy.ndarray`
            The 2D array from which to detect sources. Typically, this
            array should be an image that has been convolved with a
            smoothing kernel.

        threshold : 2D `~numpy.ndarray` or float
            The data value or pixel-wise data values (as an array) to be
            used as the per-pixel detection threshold. If ``data`` is
            a `~astropy.units.Quantity` array, then ``threshold`` must
            have the same units as ``data``. A 2D ``threshold`` array must
            have the same shape as ``data``.

        mask : 2D bool `~numpy.ndarray`, optional
            A boolean mask with the same shape as ``data``, where a
            `True` value indicates the corresponding element of ``data``
            is masked. Masked pixels will not be included in any source.

        Returns
        -------
        segment_image : `~photutils.segmentation.SegmentationImage` or `None`
            A 2D segmentation image, with the same shape as the input data,
            where sources are marked by different positive integer values. A
            value of zero is reserved for the background. If no sources are
            found then `None` is returned.
        """
        segment_img = detect_sources(data, threshold, self.n_pixels[0],
                                     mask=mask, connectivity=self.connectivity)
        if segment_img is None:
            return None

        # Source deblending requires scikit-image
        if self.deblend:
            segment_img = deblend_sources(data, segment_img, self.n_pixels[1],
                                          n_levels=self.n_levels,
                                          contrast=self.contrast,
                                          mode=self.mode,
                                          connectivity=self.connectivity,
                                          relabel=self.relabel,
                                          n_processes=self.n_processes,
                                          progress_bar=self.progress_bar)

        return segment_img



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the gridded_models module.
"""

import os.path as op
from itertools import product

import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.nddata import NDData
from astropy.table import QTable
from numpy.testing import assert_allclose, assert_equal

from photutils.datasets import make_model_image
from photutils.psf import GriddedPSFModel, STDPSFGrid
from photutils.segmentation import SourceCatalog, detect_sources
from photutils.utils._optional_deps import HAS_MATPLOTLIB

# the first file has a single detector, the rest have multiple detectors
STDPSF_FILENAMES = ('STDPSF_NRCA1_F150W_mock.fits',
                    'STDPSF_ACSWFC_F814W_mock.fits',
                    'STDPSF_NRCSW_F150W_mock.fits',
                    'STDPSF_WFC3UV_F814W_mock.fits',
                    'STDPSF_WFPC2_F814W_mock.fits')

WEBBPSF_FILENAMES = ('nircam_nrca1_f200w_fovp101_samp4_npsf16_mock.fits',
                     'nircam_nrca1_f200w_fovp101_samp4_npsf4_mock.fits',
                     'nircam_nrca5_f444w_fovp101_samp4_npsf4_mock.fits',
                     'nircam_nrcb4_f150w_fovp101_samp4_npsf1_mock.fits')


@pytest.fixture(name='psfmodel')
def fixture_griddedpsf_data():
    psfs = []
    yy, xx = np.mgrid[0:101, 0:101]
    for i in range(16):
        theta = np.deg2rad(i * 10.0)
        gmodel = Gaussian2D(1, 50, 50, 10, 5, theta=theta)
        psfs.append(gmodel(xx, yy))

    xgrid = [0, 40, 160, 200]
    ygrid = [0, 60, 140, 200]
    meta = {}
    meta['grid_xypos'] = list(product(xgrid, ygrid))
    meta['oversampling'] = 4

    nddata = NDData(psfs, meta=meta)
    return GriddedPSFModel(nddata)


class TestGriddedPSFModel:
    """
    Tests for GriddPSFModel.
    """

    def test_gridded_psf_model(self, psfmodel):
        keys = ['grid_xypos', 'oversampling']
        for key in keys:
            assert key in psfmodel.meta
        grid_xypos = psfmodel.grid_xypos
        assert len(grid_xypos) == 16
        assert_equal(psfmodel.oversampling, [4, 4])
        assert_equal(psfmodel.meta['oversampling'], psfmodel.oversampling)
        assert psfmodel.data.shape == (16, 101, 101)

        idx = np.lexsort((grid_xypos[:, 0], grid_xypos[:, 1]))
        xypos = grid_xypos[idx]
        assert_allclose(xypos, grid_xypos)

        # check that data and grid_xypos attributes are read-only
        match = 'object has no setter'
        with pytest.raises(AttributeError, match=match):
            psfmodel.data = np.ones((4, 5, 5))
        with pytest.raises(AttributeError, match=match):
            psfmodel.grid_xypos = [[0, 0], [1, 1]]

    def test_repr_str(self, psfmodel):
        repr_str = repr(psfmodel)
        assert 'GriddedPSFModel' in repr_str
        assert 'flux=1.' in repr_str
        assert 'x_0=0.' in repr_str
        assert 'y_0=0.' in repr_str
        assert 'oversampling=' in repr_str
        assert 'fill_value=0.0' in repr_str

        str_str = str(psfmodel)
        assert 'GriddedPSFModel' in str_str
        assert 'Number of PSFs: 16' in str_str
        assert 'PSF shape (oversampled pixels): (101, 101)' in str_str
        assert 'Oversampling: [4, 4]' in str_str
        assert 'Fill Value: 0.0' in str_str

    def test_gridded_psf_model_basic_eval(self, psfmodel):
        assert psfmodel(0, 0) == 1
        assert psfmodel(100, 100) == 0
        assert_allclose(psfmodel([0, 100], [0, 100]), [1, 0])

        y, x = np.mgrid[0:100, 0:100]
        psf = psfmodel.evaluate(x=x, y=y, flux=100, x_0=40, y_0=60)
        assert psf.shape == (100, 100)

        _, y2, x2 = np.mgrid[0:100, 0:100, 0:100]
        match = 'x and y must be 1D or 2D'
        with pytest.raises(ValueError, match=match):
            psfmodel.evaluate(x=x2, y=y2, flux=100, x_0=40, y_0=60)

    def test_gridded_psf_model_single_psf(self, psfmodel):
        psfmodel = psfmodel.copy()
        psfmodel._data = psfmodel.data[0:1, :, :]
        assert psfmodel(0, 0) == 1
        assert psfmodel(100, 100) == 0
        assert_allclose(psfmodel([0, 100], [0, 100]), [1, 0])

        y, x = np.mgrid[0:100, 0:100]
        psf = psfmodel.evaluate(x=x, y=y, flux=100, x_0=40, y_0=60)
        assert psf.shape == (100, 100)

        _, y2, x2 = np.mgrid[0:100, 0:100, 0:100]
        match = 'x and y must be 1D or 2D'
        with pytest.raises(ValueError, match=match):
            psfmodel.evaluate(x=x2, y=y2, flux=100, x_0=40, y_0=60)

    def test_gridded_psf_model_eval_outside_grid(self, psfmodel):
        y, x = np.mgrid[-50:50, -50:50]
        psf1 = psfmodel.evaluate(x=x, y=y, flux=100, x_0=0, y_0=0)
        y, x = np.mgrid[-60:40, -60:40]
        psf2 = psfmodel.evaluate(x=x, y=y, flux=100, x_0=-10, y_0=-10)
        assert_allclose(psf1, psf2)

        y, x = np.mgrid[150:250, 150:250]
        psf3 = psfmodel.evaluate(x=x, y=y, flux=100, x_0=200, y_0=200)
        y, x = np.mgrid[170:270, 170:270]
        psf4 = psfmodel.evaluate(x=x, y=y, flux=100, x_0=220, y_0=220)
        assert_allclose(psf3, psf4)

    def test_gridded_psf_model_invalid_inputs(self):
        data = np.ones((4, 5, 5))

        # check if NDData
        match = 'data must be an NDData instance'
        with pytest.raises(TypeError, match=match):
            GriddedPSFModel(data)

        # check PSF data dimension
        match = 'The NDData data attribute must be a 3D numpy ndarray'
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel(NDData(np.ones((3, 3))))

        match = 'The length of the PSF x and y axes must both be at least 4'
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel(NDData(np.ones((4, 3, 3))))

        match = 'The number of ePSFs must not be 2 or 3'
        meta = {'grid_xypos': [[0, 0], [1, 0], [1, 0]], 'oversampling': 4}
        nddata = NDData(np.ones((3, 4, 4)), meta=meta)
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel(nddata)

        match = 'All elements of input data must be finite'
        data2 = np.ones((4, 5, 5))
        data2[0, 2, 2] = np.nan
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel(NDData(data2))

        # check that grid_xypos is in meta
        meta = {'oversampling': 4}
        nddata = NDData(data, meta=meta)
        match = "'grid_xypos' must be in the nddata meta dictionary"
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel(nddata)

        # check grid_xypos length
        meta = {'grid_xypos': [[0, 0], [1, 0], [1, 0]], 'oversampling': 4}
        nddata = NDData(data, meta=meta)
        match = 'length of grid_xypos must match the number of input ePSFs'
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel(nddata)

        # check if grid_xypos is a regular grid
        meta = {'grid_xypos': [[0, 0], [1, 0], [1, 0], [3, 4]],
                'oversampling': 4}
        nddata = NDData(data, meta=meta)
        match = 'grid_xypos must form a rectangular grid'
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel(nddata)

        meta = {'grid_xypos': [[0, 0], [0, 2], [0, 4], [0, 6]],
                'oversampling': 4}
        nddata = NDData(data, meta=meta)
        match = 'grid_xypos must form a rectangular grid'
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel(nddata)

        # check that oversampling is in meta
        meta = {'grid_xypos': [[0, 0], [0, 1], [1, 0], [1, 1]]}
        nddata = NDData(data, meta=meta)
        match = "'oversampling' must be in the nddata meta dictionary"
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel(nddata)

    def test_gridded_psf_model_eval(self, psfmodel):
        """
        Create a simulated image using GriddedPSFModel and test the
        properties of the generated sources.
        """
        shape = (200, 200)
        params = QTable()
        params['x_0'] = [40, 50, 160, 160]
        params['y_0'] = [60, 150, 50, 140]
        params['flux'] = [100, 100, 100, 100]
        data = make_model_image(shape, psfmodel, params)

        segm = detect_sources(data, 0.0, 5)
        cat = SourceCatalog(data, segm)
        orients = cat.orientation.value
        assert_allclose(orients[1], 50.0, rtol=1.0e-5)
        assert_allclose(orients[2], 280.0, rtol=1.0e-5)
        assert 88.3 < orients[0] < 88.4
        assert 64.0 < orients[3] < 64.2

    @pytest.mark.parametrize('deepcopy', [False, True])
    def test_copy(self, psfmodel, deepcopy):
        flux = psfmodel.flux.value
        model_copy = psfmodel.deepcopy() if deepcopy else psfmodel.copy()

        assert_equal(model_copy.data, psfmodel.data)
        assert_equal(model_copy.grid_xypos, psfmodel.grid_xypos)
        assert_equal(model_copy.oversampling, psfmodel.oversampling)
        assert_equal(model_copy.meta, psfmodel.meta)
        assert model_copy.flux.value == psfmodel.flux.value
        assert model_copy.x_0.value == psfmodel.x_0.value
        assert model_copy.y_0.value == psfmodel.y_0.value
        assert model_copy.fixed == psfmodel.fixed

        model_copy.data[0, 0, 0] = 42
        if deepcopy:
            assert model_copy.data[0, 0, 0] != psfmodel.data[0, 0, 0]
        else:
            assert model_copy.data[0, 0, 0] == psfmodel.data[0, 0, 0]

        model_copy.flux = 100
        assert model_copy.flux.value != flux

        model_copy.x_0.fixed = True
        model_copy.y_0.fixed = True
        new_model = model_copy.copy()
        assert new_model.x_0.fixed
        assert new_model.fixed == model_copy.fixed

    def test_repr(self, psfmodel):
        model_repr = repr(psfmodel)
        assert '<GriddedPSFModel(' in model_repr
        for param in psfmodel.param_names:
            assert param in model_repr

    def test_str(self, psfmodel):
        model_str = str(psfmodel)
        keys = ('Grid shape', 'Number of PSFs', 'PSF shape', 'Oversampling')
        for key in keys:
            assert key in model_str
        for param in psfmodel.param_names:
            assert param in model_str

    def test_gridded_psf_oversampling(self, psfmodel):
        nddata = NDData(psfmodel.data, meta=psfmodel.meta)
        nddata.meta['oversampling'] = [4, 4]
        psfmodel2 = GriddedPSFModel(nddata)
        assert_equal(psfmodel2.oversampling, psfmodel.oversampling)

    def test_bounding_box(self, psfmodel):
        # oversampling is 4
        bbox = psfmodel.bounding_box.bounding_box()
        assert_equal(bbox, ((-12.625, 12.625), (-12.625, 12.625)))

        model = psfmodel.copy()
        model.oversampling = 1
        bbox = model.bounding_box.bounding_box()
        assert_equal(bbox, ((-50.5, 50.5), (-50.5, 50.5)))

    def test_read_stdpsf(self):
        """
        Test STDPSF read for a single detector.
        """
        filename = 'STDPSF_NRCA1_F150W_mock.fits'
        filename = op.join(op.dirname(op.abspath(__file__)), 'data', filename)
        psfmodel = GriddedPSFModel.read(filename)
        assert psfmodel.data.shape[0] == len(psfmodel.meta['grid_xypos'])
        assert_equal(psfmodel.oversampling, [4, 4])
        assert_equal(psfmodel.meta['oversampling'], psfmodel.oversampling)

    @pytest.mark.parametrize('filename', STDPSF_FILENAMES[1:])
    @pytest.mark.parametrize('detector_id', [1, 2])
    def test_read_stdpsf_multi_detector(self, filename, detector_id):
        """
        Test STDPSF read for multiple detectors.
        """
        filename = op.join(op.dirname(op.abspath(__file__)), 'data', filename)
        psfmodel = GriddedPSFModel.read(filename, detector_id=detector_id,
                                        format='stdpsf')
        assert psfmodel.data.shape[0] == len(psfmodel.meta['grid_xypos'])
        assert_equal(psfmodel.oversampling, [4, 4])
        assert_equal(psfmodel.meta['oversampling'], psfmodel.oversampling)

        # test format auto-detect
        filename = op.join(op.dirname(op.abspath(__file__)), 'data', filename)
        psfmodel = GriddedPSFModel.read(filename, detector_id=detector_id)
        assert psfmodel.data.shape[0] == len(psfmodel.meta['grid_xypos'])

        match = 'detector_id must be specified'
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel.read(filename, detector_id=None)

        match = 'detector_id must be '
        with pytest.raises(ValueError, match=match):
            GriddedPSFModel.read(filename, detector_id=-1)

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    @pytest.mark.parametrize('filename', WEBBPSF_FILENAMES)
    def test_read_webbpsf(self, filename):
        filename = op.join(op.dirname(op.abspath(__file__)), 'data', filename)
        psfmodel = GriddedPSFModel.read(filename, format='webbpsf')
        assert psfmodel.data.shape[0] == len(psfmodel.meta['grid_xypos'])
        assert_equal(psfmodel.oversampling, [4, 4])
        assert_equal(psfmodel.meta['oversampling'], psfmodel.oversampling)
        psfmodel.plot_grid()

        # test format auto-detect
        filename = op.join(op.dirname(op.abspath(__file__)), 'data', filename)
        psfmodel = GriddedPSFModel.read(filename)
        assert psfmodel.data.shape[0] == len(psfmodel.meta['grid_xypos'])

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
    def test_plot(self, psfmodel):
        psfmodel.plot_grid()
        psfmodel.plot_grid(peak_norm=True, cmap='Blues', vmax_scale=0.9)
        psfmodel.plot_grid(deltas=True)
        psfmodel.plot_grid(deltas=True, peak_norm=True)

        # simulate a grid where one or more ePSFS are blank (all zeros)
        model = psfmodel.deepcopy()
        model.data[0] = 0.0
        model.plot_grid(deltas=True)


@pytest.mark.skipif(not HAS_MATPLOTLIB, reason='matplotlib is required')
@pytest.mark.parametrize('filename', STDPSF_FILENAMES)
def test_stdpsfgrid(filename):
    filename = op.join(op.dirname(op.abspath(__file__)), 'data', filename)
    psfgrid = STDPSFGrid(filename)
    assert 'grid_xypos' in psfgrid.meta
    assert 'oversampling' in psfgrid.meta
    assert_equal(psfgrid.oversampling, [4, 4])
    assert psfgrid.data.shape[0] == len(psfgrid.meta['grid_xypos'])

    psfgrid.plot_grid()


def test_stdpsfgrid_repr_str():
    filename = STDPSF_FILENAMES[0]
    filename = op.join(op.dirname(op.abspath(__file__)), 'data', filename)
    psfgrid = STDPSFGrid(filename)
    assert repr(psfgrid) == str(psfgrid)
    keys = ('STDPSF', 'Grid_shape', 'Number of PSFs', 'PSF shape',
            'Oversampling')
    for key in keys:
        assert key in repr(psfgrid)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the interpolation module.
"""

from unittest.mock import MagicMock

import numpy as np
import pytest
from numpy.testing import assert_allclose

from photutils.utils import ShepardIDWInterpolator as IDWInterp

SHAPE = (5, 5)
DATA = np.ones(SHAPE) * 2.0
MASK = np.zeros(DATA.shape, dtype=bool)
MASK[2, 2] = True
ERROR = np.ones(SHAPE)
BACKGROUND = np.ones(SHAPE)
WRONG_SHAPE = np.ones((2, 2))


class TestShepardIDWInterpolator:
    def setup_class(self):
        self.rng = np.random.default_rng(0)
        self.x = self.rng.random(100)
        self.y = np.sin(self.x)
        self.f = IDWInterp(self.x, self.y)

    @pytest.mark.parametrize('positions', [0.4, np.arange(2, 5) * 0.1])
    def test_idw_1d(self, positions):
        """
        Test 1D IDW interpolation.
        """
        f = IDWInterp(self.x, self.y)
        assert_allclose(f(positions), np.sin(positions), atol=1e-2)

    def test_idw_weights(self):
        """
        Test IDW interpolation with weights.
        """
        weights = self.y * 0.1
        f = IDWInterp(self.x, self.y, weights=weights)
        pos = 0.4
        assert_allclose(f(pos), np.sin(pos), atol=1e-2)

    def test_idw_2d(self):
        """
        Test 2D IDW interpolation.
        """
        pos = self.rng.random((1000, 2))
        val = np.sin(pos[:, 0] + pos[:, 1])
        f = IDWInterp(pos, val)
        x = 0.5
        y = 0.6
        assert_allclose(f([x, y]), np.sin(x + y), atol=1e-2)

    def test_idw_3d(self):
        """
        Test 3D IDW interpolation.
        """
        val = np.ones((3, 3, 3))
        pos = np.indices(val.shape)
        f = IDWInterp(pos, val)
        assert_allclose(f([0.5, 0.5, 0.5]), 1.0)

    def test_no_coordinates(self):
        """
        Test that empty coordinates raises ValueError.
        """
        match = 'coordinates must have at least one data point'
        with pytest.raises(ValueError, match=match):
            IDWInterp([], 0)

    def test_values_invalid_shape(self):
        """
        Test that mismatched values shape raises ValueError.
        """
        match = 'The number of values must match the number of coordinates'
        with pytest.raises(ValueError, match=match):
            IDWInterp(self.x, 0)

    def test_weights_invalid_shape(self):
        """
        Test that mismatched weights shape raises ValueError.
        """
        match = 'number of weights must match the number of coordinates'
        with pytest.raises(ValueError, match=match):
            IDWInterp(self.x, self.y, weights=10)

    def test_weights_negative(self):
        """
        Test that negative weights raises ValueError.
        """
        match = 'All weight values must be non-negative numbers'
        with pytest.raises(ValueError, match=match):
            IDWInterp(self.x, self.y, weights=-self.y)

    def test_n_neighbors_one(self):
        """
        Test IDW interpolation with n_neighbors=1.
        """
        result = self.f(0.5, n_neighbors=1)
        assert np.isscalar(result)
        assert_allclose(result, 0.479334, rtol=3e-7)

    def test_n_neighbors_negative(self):
        """
        Test that negative n_neighbors raises ValueError.
        """
        match = 'n_neighbors must be a positive integer'
        with pytest.raises(ValueError, match=match):
            self.f(0.5, n_neighbors=-1)

    def test_conf_dist_negative(self):
        """
        Test IDW interpolation with negative conf_dist.
        """
        assert_allclose(self.f(0.5, conf_dist=-1),
                        self.f(0.5, conf_dist=None))

    def test_dtype_none(self):
        """
        Test IDW interpolation with dtype=None.
        """
        result = self.f(0.5, dtype=None)
        assert result.dtype == float

    def test_positions_0d_nomatch(self):
        """
        Test that a 0D position with wrong dimensionality raises
        ValueError.
        """
        pos = self.rng.random((10, 2))
        val = np.sin(pos[:, 0] + pos[:, 1])
        f = IDWInterp(pos, val)
        match = 'position does not match the dimensionality'
        with pytest.raises(ValueError, match=match):
            f(0.5)

    def test_positions_1d_nomatch(self):
        """
        Test that a 1D position with wrong length raises ValueError.
        """
        pos = self.rng.random((10, 2))
        val = np.sin(pos[:, 0] + pos[:, 1])
        f = IDWInterp(pos, val)
        match = 'was provided as a 1D array, but its length does not match'
        with pytest.raises(ValueError, match=match):
            f([0.5])

    def test_positions_3d(self):
        """
        Test that a 3D positions array raises ValueError.
        """
        match = 'array_like object of dimensionality no larger than 2'
        with pytest.raises(ValueError, match=match):
            self.f(np.ones((3, 3, 3)))

    def test_scalar_values_1d(self):
        """
        Test IDW interpolation with a single 1D data point.
        """
        value = 10.0
        f = IDWInterp(2, value)
        assert_allclose(f(2), value)
        assert_allclose(f(-1), value)
        assert_allclose(f(0), value)
        assert_allclose(f(142), value)

    def test_scalar_values_2d(self):
        """
        Test IDW interpolation with a single 2D data point.
        """
        value = 10.0
        f = IDWInterp([[1, 2]], value)
        assert_allclose(f([1, 2]), value)
        assert_allclose(f([-1, 0]), value)
        assert_allclose(f([142, 213]), value)

    def test_scalar_values_3d(self):
        """
        Test IDW interpolation with a single 3D data point.
        """
        value = 10.0
        f = IDWInterp([[7, 4, 1]], value)
        assert_allclose(f([7, 4, 1]), value)
        assert_allclose(f([-1, 0, 7]), value)
        assert_allclose(f([142, 213, 5]), value)

    def test_no_valid_distances(self):
        """
        Test that NaN is returned when all distances are non-finite
        (dk.shape[0] == 0 after filtering).
        """
        coords = np.array([0.0, 1.0, 2.0])
        values = np.array([10.0, 20.0, 30.0])
        f = IDWInterp(coords, values)

        # Replace the kdtree with a mock that returns all-inf distances
        mock_kdtree = MagicMock()
        inf_distances = np.array([[np.inf, np.inf, np.inf]])
        inf_idx = np.array([[0, 1, 2]])
        mock_kdtree.query.return_value = (inf_distances, inf_idx)
        f.kdtree = mock_kdtree

        result = f(0.5, n_neighbors=3)
        assert np.isnan(result)

    def test_zero_weights(self):
        """
        Test that NaN is returned when all weights are zero, leading
        to wtot == 0.
        """
        coords = np.array([0.0, 1.0, 2.0])
        values = np.array([10.0, 20.0, 30.0])
        weights = np.array([0.0, 0.0, 0.0])
        f = IDWInterp(coords, values, weights=weights)
        result = f(0.5, n_neighbors=3)
        assert np.isnan(result)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for convolving images with a kernel.
"""

import warnings

import numpy as np
from astropy.convolution import Kernel2D
from astropy.units import Quantity
from astropy.utils.exceptions import AstropyUserWarning
from scipy.ndimage import convolve as ndi_convolve


def _filter_data(data, kernel, *, mode='constant', fill_value=0.0,
                 check_normalization=False):
    """
    Convolve a 2D image with a 2D kernel.

    The kernel may either be a 2D `~numpy.ndarray` or a
    `~astropy.convolution.Kernel2D` object.

    Parameters
    ----------
    data : array_like
        The 2D array of the image.

    kernel : array_like (2D) or `~astropy.convolution.Kernel2D`
        The 2D kernel used to filter the input ``data``. Filtering the
        ``data`` will smooth the noise and maximize detectability of
        objects with a shape similar to the kernel.

    mode : {'constant', 'reflect', 'nearest', 'mirror', 'wrap'}, optional
        The ``mode`` determines how the array borders are handled. For
        the ``'constant'`` mode, values outside the array borders are
        set to ``fill_value``. The default is ``'constant'``.

    fill_value : scalar, optional
        Value to fill data values beyond the array borders if ``mode``
        is ``'constant'``. The default is ``0.0``. When ``data`` is a
        `~astropy.units.Quantity`, the result has the same unit; the
        numerical value of ``fill_value`` is used as-is (it is not
        converted to the data unit).

    check_normalization : bool, optional
        If `True` then a warning will be issued if the kernel is not
        normalized to 1.

    Returns
    -------
    result : `~numpy.ndarray` or `~astropy.units.Quantity`
        The convolved image. A `~astropy.units.Quantity` is returned if
        ``data`` has units; otherwise a `~numpy.ndarray`.
    """
    if kernel is None:
        return data

    kernel_array = kernel.array if isinstance(kernel, Kernel2D) else kernel

    if check_normalization and not np.allclose(np.sum(kernel_array), 1.0):
        msg = 'The kernel is not normalized.'
        warnings.warn(msg, AstropyUserWarning)

    # scipy.ndimage.convolve currently strips units, but be explicit in
    # case that behavior changes
    unit = None
    if isinstance(data, Quantity):
        unit = data.unit
        data = data.value

    # NOTE: if data is int and kernel is float, ndimage.convolve will
    # return an int image. If the data dtype is int, we make the data
    # float so that a float image is always returned
    if np.issubdtype(data.dtype, np.integer):
        data = data.astype(float)

    # NOTE: astropy.convolution.convolve fails with zero-sum kernels
    # (used in findstars) (cf. astropy #1647)
    result = ndi_convolve(data, kernel_array, mode=mode, cval=fill_value)

    # Reapply the input unit
    if unit is not None:
        result <<= unit

    return result



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Subpackage containing general-purpose utility functions that do not fit
into any of the other subpackages.
"""

from .colormaps import *  # noqa: F401, F403
from .cutouts import *  # noqa: F401, F403
from .depths import *  # noqa: F401, F403
from .errors import *  # noqa: F401, F403
from .exceptions import *  # noqa: F401, F403
from .footprints import *  # noqa: F401, F403
from .interpolation import *  # noqa: F401, F403



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the core module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from numpy.testing import assert_allclose

from photutils.aperture import (Aperture, CircularAperture, EllipticalAperture,
                                SkyCircularAperture)
from photutils.aperture.core import _aperture_metadata

POSITIONS = [(5, 5), (10, 10), (15, 15)]
SCALAR_POS = (5, 5)


class MinimalAperture(Aperture):
    """
    Minimal concrete Aperture subclass that is neither PixelAperture nor
    SkyAperture.

    Used to exercise bare-Aperture code paths.
    """

    _params = ('positions',)

    @property
    def positions(self):
        """
        Return a fixed single position.
        """
        return np.array([[5.0, 5.0]])


class RaisesOnCompare:
    """
    Helper object that raises TypeError on any != comparison, used to
    exercise the except-TypeError branch in Aperture.__eq__.
    """

    def __ne__(self, other):
        """
        Raise TypeError unconditionally.
        """
        msg = 'incompatible types'
        raise TypeError(msg)


class TestAperture:
    """
    Tests for branches of the Aperture base class not covered elsewhere.
    """

    def test_positions_str_raises_for_unknown_type(self):
        """
        Test that _positions_str raises TypeError when the aperture is
        not a PixelAperture or SkyAperture subclass.
        """
        aper = MinimalAperture()
        match = 'Aperture must be a subclass of PixelAperture or SkyAperture'
        with pytest.raises(TypeError, match=match):
            aper._positions_str()

    def test_eq_different_class(self):
        """
        Test that __eq__ returns False when compared to a different
        class (exercises the isinstance early-return branch).
        """
        aper1 = CircularAperture(SCALAR_POS, r=3)
        aper2 = EllipticalAperture(SCALAR_POS, a=3, b=2, theta=0)
        assert aper1 != aper2

    def test_eq_different_params(self):
        """
        Test that __eq__ returns False when the two instances have
        different _params tuples (exercises the params-mismatch branch).
        """
        aper1 = CircularAperture(SCALAR_POS, r=3)
        aper2 = CircularAperture(SCALAR_POS, r=3)
        # Inject an extended _params tuple at the instance level so that
        # isinstance passes (both CircularAperture instances, same type so
        # Python does not invoke subclass reflection) while the params
        # check diverges, hitting the mismatch return at line 104.
        aper2.__dict__['_params'] = (*CircularAperture._params, 'extra')
        assert aper1 != aper2

    def test_eq_comparison_type_error(self):
        """
        Test that __eq__ returns False (rather than propagating the
        exception) when the position comparison raises TypeError.
        """
        aper1 = CircularAperture(SCALAR_POS, r=3)
        aper2 = CircularAperture(SCALAR_POS, r=3)
        # Bypass the descriptor and inject a position object that raises
        # TypeError on != comparison, mimicking incompatible SkyCoords.
        aper1.__dict__['positions'] = RaisesOnCompare()
        aper2.__dict__['positions'] = RaisesOnCompare()
        assert aper1 != aper2


class TestPixelAperture:
    """
    Tests for branches of the PixelAperture class not covered elsewhere.
    """

    def test_to_mask_invalid_method(self):
        """
        Test that to_mask raises ValueError for an unrecognised
        method string (exercises the invalid-method branch in
        _translate_mask_method).
        """
        aper = CircularAperture(SCALAR_POS, r=3)
        match = 'Invalid mask method'
        with pytest.raises(ValueError, match=match):
            aper.to_mask(method='invalid')

    def test_bbox_multi_position(self):
        """
        Test that the bbox property returns a list for a multi-position
        aperture (exercises the non-scalar branch).
        """
        aper = CircularAperture(POSITIONS, r=3)
        bbox = aper.bbox
        assert isinstance(bbox, list)
        assert len(bbox) == len(POSITIONS)


class TestPixelApertureDoPhotometry:
    """
    Tests for error-handling branches of PixelAperture.do_photometry.
    """

    def setup_method(self):
        """
        Set up a simple scalar aperture and matching data array.
        """
        self.aper = CircularAperture(SCALAR_POS, r=3)
        self.data = np.ones((20, 20))

    def test_do_photometry_1d_data_error(self):
        """
        Test that do_photometry raises ValueError when data is not a
        2D array.
        """
        match = 'data must be a 2D array'
        with pytest.raises(ValueError, match=match):
            self.aper.do_photometry(np.ones(20))

    def test_do_photometry_error_shape_mismatch(self):
        """
        Test that do_photometry raises ValueError when the error array
        does not match the data shape.
        """
        match = 'error and data must have the same shape'
        with pytest.raises(ValueError, match=match):
            self.aper.do_photometry(self.data, error=np.ones((5, 5)))

    def test_do_photometry_unit_mismatch_error(self):
        """
        Test that do_photometry raises ValueError when data and error
        have different units.
        """
        import astropy.units as u

        match = 'they both must have the same units'
        with pytest.raises(ValueError, match=match):
            self.aper.do_photometry(
                self.data * u.Jy,
                error=self.data * u.ct,
            )

    def test_do_photometry_basic(self):
        """
        Test that do_photometry returns the expected aperture sum for a
        uniform data array with no error input.
        """
        sums, errs = self.aper.do_photometry(self.data)
        assert_allclose(sums[0], np.pi * 9, rtol=1e-3)
        assert len(errs) == 0


class TestApertureReprStr:
    """
    Tests for __repr__ and __str__ of various aperture types.
    """

    def test_repr_scalar(self):
        """
        Test __repr__ for a scalar CircularAperture.
        """
        aper = CircularAperture(SCALAR_POS, r=3)
        result = repr(aper)
        assert 'CircularAperture' in result
        assert 'r=3.0' in result

    def test_repr_multi(self):
        """
        Test __repr__ for a multi-position CircularAperture.
        """
        aper = CircularAperture(POSITIONS, r=3)
        result = repr(aper)
        assert 'CircularAperture' in result

    def test_str_scalar(self):
        """
        Test __str__ for a scalar CircularAperture.
        """
        aper = CircularAperture(SCALAR_POS, r=3)
        result = str(aper)
        assert 'Aperture: CircularAperture' in result
        assert 'r: 3.0' in result

    def test_str_multi(self):
        """
        Test __str__ for a multi-position CircularAperture.
        """
        aper = CircularAperture(POSITIONS, r=3)
        result = str(aper)
        assert 'Aperture: CircularAperture' in result

    def test_repr_sky(self):
        """
        Test __repr__ for a SkyCircularAperture.
        """
        pos = SkyCoord(ra=10, dec=20, unit='deg')
        aper = SkyCircularAperture(pos, r=1.0 * u.arcsec)
        result = repr(aper)
        assert 'SkyCircularAperture' in result


class TestApertureIteration:
    """
    Tests for __iter__ and __len__ of Aperture objects.
    """

    def test_iter(self):
        """
        Test that iterating over a multi-position aperture yields
        scalar apertures.
        """
        aper = CircularAperture(POSITIONS, r=3)
        items = list(aper)
        assert len(items) == len(POSITIONS)
        for item in items:
            assert item.isscalar

    def test_copy(self):
        """
        Test that copy creates a deep copy with independent data.
        """
        aper = CircularAperture(POSITIONS, r=3)
        aper_copy = aper.copy()
        assert aper == aper_copy
        aper_copy.r = 5.0
        assert aper != aper_copy
        assert aper.r == 3.0

    def test_copy_sky(self):
        """
        Test that copy works for SkyAperture objects.
        """
        pos = SkyCoord(ra=[10, 20], dec=[30, 40], unit='deg')
        aper = SkyCircularAperture(pos, r=1.0 * u.arcsec)
        aper_copy = aper.copy()
        assert aper == aper_copy


class TestApertureMetadata:
    """
    Tests for the _aperture_metadata helper function.
    """

    def test_metadata_keys(self):
        """
        Test that _aperture_metadata returns the expected keys.
        """
        aper = CircularAperture(SCALAR_POS, r=3)
        meta = _aperture_metadata(aper)
        assert 'aperture' in meta
        assert meta['aperture'] == 'CircularAperture'
        assert 'aperture_r' in meta
        assert meta['aperture_r'] == 3.0

    def test_metadata_with_index(self):
        """
        Test that _aperture_metadata uses the index in keys.
        """
        aper = CircularAperture(SCALAR_POS, r=3)
        meta = _aperture_metadata(aper, index='_0')
        assert 'aperture_0' in meta
        assert meta['aperture_0'] == 'CircularAperture'
        assert 'aperture_0_r' in meta

    def test_metadata_class_name_not_repeated(self):
        """
        Test that aperture class name key is set exactly once.
        """
        aper = EllipticalAperture(SCALAR_POS, a=5, b=3, theta=0)
        meta = _aperture_metadata(aper)
        assert meta['aperture'] == 'EllipticalAperture'
        assert 'aperture_a' in meta
        assert 'aperture_b' in meta
        assert 'aperture_theta' in meta



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for creating images from PSF models.
"""

import numpy as np

from photutils.datasets import make_model_image, make_model_params
from photutils.datasets.images import _model_shape_from_bbox
from photutils.psf.utils import _get_psf_model_main_params
from photutils.utils._parameters import as_pair

__all__ = ['make_psf_model_image']


def make_psf_model_image(shape, psf_model, n_sources, *, model_shape=None,
                         min_separation=1, border_size=None, seed=0,
                         progress_bar=False, **kwargs):
    """
    Make an example image containing PSF model images.

    Source parameters are randomly generated using an optional ``seed``.

    Parameters
    ----------
    shape : 2-tuple of int
        The shape of the output image.

    psf_model : 2D `astropy.modeling.Model`
        The PSF model. The model must have parameters named ``x_0``,
        ``y_0``, and ``flux``, corresponding to the center (x, y)
        position and flux, or it must have 'x_name', 'y_name', and
        'flux_name' attributes that map to the x, y, and flux parameters
        (i.e., a model output from `make_psf_model`). The model must be
        two-dimensional such that it accepts 2 inputs (e.g., x and y)
        and provides 1 output.

    n_sources : int
        The number of sources to generate. If ``min_separation`` is too
        large, the number of requested sources may not fit within the
        given ``shape`` and therefore the number of sources generated
        may be less than ``n_sources``.

    model_shape : `None` or 2-tuple of int, optional
        The shape around the center (x, y) position that will used to
        evaluate the ``psf_model``. If `None`, then the shape will be
        determined from the ``psf_model`` bounding box (an error will be
        raised if the model does not have a bounding box).

    min_separation : float, optional
        The minimum separation between the centers of two sources. Note
        that if the minimum separation is too large, the number of
        sources generated may be less than ``n_sources``.

    border_size : `None`, tuple of 2 int, or int, optional
        The (ny, nx) size of the exclusion border around the image edges
        where no sources will be generated that have centers within
        the border region. If a single integer is provided, it will be
        used for both dimensions. If `None`, then a border size equal
        to half the (y, x) size of the evaluated PSF model (taking any
        oversampling into account) will be used.

    seed : int, optional
        A seed to initialize the `numpy.random.BitGenerator`. If `None`,
        then fresh, unpredictable entropy will be pulled from the OS.

    progress_bar : bool, optional
        Whether to display a progress bar when creating the sources. The
        progress bar requires that the `tqdm <https://tqdm.github.io/>`_
        optional dependency be installed.

    **kwargs
        Keyword arguments are accepted for additional model parameters.
        The values should be 2-tuples of the lower and upper bounds for
        the parameter range. The parameter values will be uniformly
        distributed between the lower and upper bounds, inclusively. If
        the parameter is not in the input ``psf_model`` parameter names,
        it will be ignored.

    Returns
    -------
    data : 2D `~numpy.ndarray`
        The simulated image.

    table : `~astropy.table.Table`
        A table containing the (x, y, flux) parameters of the generated
        sources. The column names will correspond to the names of the
        input ``psf_model`` (x, y, flux) parameter names. The table will
        also contain an ``'id'`` column with unique source IDs.

    Examples
    --------
    >>> from photutils.psf import CircularGaussianPRF, make_psf_model_image
    >>> shape = (150, 200)
    >>> psf_model = CircularGaussianPRF(fwhm=3.5)
    >>> n_sources = 10
    >>> data, params = make_psf_model_image(shape, psf_model, n_sources,
    ...                                     flux=(100, 250),
    ...                                     min_separation=10,
    ...                                     seed=0)
    >>> params['x_0'].info.format = '.4f'  # optional format
    >>> params['y_0'].info.format = '.4f'
    >>> params['flux'].info.format = '.4f'
    >>> print(params)  # doctest: +FLOAT_CMP
     id   x_0      y_0      flux
    --- -------- -------- --------
      1 125.2010  72.3184 147.9522
      2  57.6408  39.1380 128.1262
      3  15.5391 115.4520 200.8790
      4  11.0411 131.7530 129.2661
      5 157.6417  43.6615 186.6532
      6 175.9470  80.2172 190.3359
      7 142.2274 132.7563 244.3635
      8 108.0270  13.4284 110.8398
      9 180.0533 106.0888 174.9959
     10 158.1171  90.3260 211.6146

    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from photutils.psf import CircularGaussianPRF, make_psf_model_image

        shape = (150, 200)
        psf_model = CircularGaussianPRF(fwhm=3.5)
        n_sources = 10
        data, params = make_psf_model_image(shape, psf_model, n_sources,
                                            flux=(100, 250),
                                            min_separation=10,
                                            seed=0)
        fig, ax = plt.subplots()
        ax.imshow(data, origin='lower')

    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from photutils.psf import CircularGaussianPRF, make_psf_model_image

        shape = (150, 200)
        psf_model = CircularGaussianPRF(fwhm=3.5)
        n_sources = 10
        data, params = make_psf_model_image(shape, psf_model, n_sources,
                                            flux=(100, 250),
                                            min_separation=10,
                                            seed=0, sigma=(1, 2))
        fig, ax = plt.subplots()
        ax.imshow(data, origin='lower')
    """
    main_params = _get_psf_model_main_params(psf_model)

    if model_shape is not None:
        model_shape = as_pair('model_shape', model_shape, lower_bound=(0, 0))
    else:
        try:
            model_shape = _model_shape_from_bbox(psf_model)
        except ValueError as exc:
            msg = ('model_shape must be specified if the model does not '
                   'have a bounding_box attribute')
            raise ValueError(msg) from exc

    if border_size is None:
        border_size = (np.array(model_shape) - 1) // 2

    other_params = {}
    if kwargs:
        # include only kwargs that are not x, y, or flux (main params)
        for key, val in kwargs.items():
            if key not in psf_model.param_names or key in main_params[0:2]:
                continue  # skip the x, y parameters
            other_params[key] = val

    x_name, y_name = main_params[0:2]
    params = make_model_params(shape, n_sources, x_name=x_name, y_name=y_name,
                               min_separation=min_separation,
                               border_size=border_size, seed=seed,
                               **other_params)

    data = make_model_image(shape, psf_model, params, model_shape=model_shape,
                            x_name=x_name, y_name=y_name,
                            progress_bar=progress_bar)

    return data, params



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Base classes for aperture tests.
"""

from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from numpy.testing import assert_equal


class BaseTestApertureParams:
    index = 2
    slc = slice(0, 2)
    expected_slc_len = 2


class BaseTestAperture(BaseTestApertureParams):
    def test_index(self):
        aper = self.aperture[self.index]
        assert isinstance(aper, self.aperture.__class__)
        assert aper.isscalar
        expected_positions = self.aperture.positions[self.index]

        for param in aper._params:
            if param == 'positions':
                if isinstance(expected_positions, SkyCoord):
                    assert_quantity_allclose(aper.positions.ra,
                                             expected_positions.ra)
                    assert_quantity_allclose(aper.positions.dec,
                                             expected_positions.dec)
                else:
                    assert_equal(getattr(aper, param), expected_positions)
            else:
                assert (getattr(aper, param)
                        == getattr(self.aperture, param))

    def test_slice(self):
        aper = self.aperture[self.slc]
        assert isinstance(aper, self.aperture.__class__)
        assert len(aper) == self.expected_slc_len
        expected_positions = self.aperture.positions[self.slc]

        for param in aper._params:
            if param == 'positions':
                if isinstance(expected_positions, SkyCoord):
                    assert_quantity_allclose(aper.positions.ra,
                                             expected_positions.ra)
                    assert_quantity_allclose(aper.positions.dec,
                                             expected_positions.dec)
                else:
                    assert_equal(getattr(aper, param), expected_positions)
            else:
                assert (getattr(aper, param)
                        == getattr(self.aperture, param))
